#ifndef SEPMYVECTOR_H
#define SEPMYVECTOR_H 1
#include <vector>
#include <algorithm>
#include <iterator>
#include <string>
#include <cstring>
#include <assert.h>
#include "seplib.h"
extern "C" {
#include "sep3d.h"
}
using std::string;
using std::vector;
namespace SEPMY {

// class for single axis
class axis {
public:
    axis();
    axis(int n, float o = 0.0, float d = 1.0, string label = "Undefined", string unit = "Undefined");
    void set(int n, float o = 0.0, float d = 1.0, string label = "Undefined", string unit = "Undefined");
    void axisInfo(int i = 1);
    int n;
    float o, d;
    string label, unit;
    bool axisInitialized = false;
private:
};

// class for hyper cube
// n1 is fast axis, n7 is slow axis
class hyperCube {
public:
    hyperCube() {}
    hyperCube(std::vector<axis> axes);
    void setAxes(std::vector<axis> axes);
    void setAxis(int iDim, axis myAxis);
    axis getAxis(int iDim);
    std::vector<axis> getAxes();
    virtual void writeAxes(string tag);
    virtual void readAxes(string tag);
    int getSize();
    void axesInfo();
    bool axesExist();
private:
    std::vector<axis> axes;
    bool axesInitialized = false;
    int naxis = 0;
};

class MYVector: public hyperCube{
public:
	// constructor and destructor
	MYVector();
	MYVector(int n1, int n2 = 0, int n3 = 0, int n4 = 0, int n5 = 0, int n6 = 0, int n7 = 0);
	MYVector(axis ax1, axis ax2 = axis(), axis ax3 = axis(), axis ax4 = axis(), axis ax5 = axis(), axis ax6 = axis(), axis ax7 = axis());
	MYVector(std::vector<axis> axes);
	~MYVector();
	void allocateMemory();
	 // file IO
	 void readFromFile(string tag);
	 void writeToFile(string tag);

	 // data reference
	 float& operator()(int i1, int i2 = 0, int i3 = 0, int i4 = 0, int i5 = 0, int i6 = 0, int i7 = 0);
	 float& operator[](int i);
	 float* getData() const { return dataPtr;}
	 int size() const {return nsize;}
	 // linear algebra opetator
	 void clear();

private:
	//temperaty public
	void init(std::vector<axis> axes);
	int capacity;
	int nsize;
	int ndim;
	std::vector<int> nvec;
	std::vector<int> unitvec;
	std::vector<float> ovec;
	std::vector<float> dvec;
	float * dataPtr = NULL;
};

// Declarition of linear algebra operator
//CLEAR
void clear(MYVector& vec);
double innerProd(MYVector& x, MYVector& y);
void addInv(MYVector& x);
// y = a * x + b * y
void linearCombo(float a, const MYVector& x, float b, MYVector& y);


//	virtual void linearCombo(float a, numCube  * x, float b, numCube * y) const;


//	virtual void addInv(numCube  * x, dataContainer * xNeg) const ;
//	virtual void scale(float a, numCube * x)const ;
//	virtual void scale(float a, numCube * x,dataContainer * ax)const ;
//	virtual void random(numCube * x) const ;
//	virtual ~numCube(){ }




// Below is actual implementation for axis class
axis::axis(){
	n = 0; o = 0; d = 0; label = "Undefined"; unit = "Undefined";
}
axis::axis(int n, float o, float d, string label, string unit) {
    set(n, o, d, label, unit);
}

void axis::axisInfo(int i) {
    fprintf(stderr, "Axis %d details: n = %d, o = %f, d = %f, ", i, (int)n, o, d);
    std::cout << "label = " << label << ", unit = " << unit << std::endl;
}

void axis::set(int n, float o, float d, string label, string unit) {
    this->n = n;
    this->o = o;
    this->d = d;
    this->label = label;
    this->unit = unit;
    axisInitialized = true;
}

// Below is the actual implementation for hypercube class
hyperCube::hyperCube(std::vector<axis> axes) {
	setAxes(axes);
}

void hyperCube::setAxes(std::vector<axis> axes) {
	this->axes = axes;
	axesInitialized = true;
	naxis = (int)axes.size();
}

void hyperCube::setAxis(int iDim, axis myAxis) {
	assert(axesInitialized && iDim >= 1 && iDim <= naxis);
	axes[iDim - 1] = myAxis;
}

axis hyperCube::getAxis(int iDim) {
	assert(axesInitialized && iDim >= 1 && iDim <= naxis);
	return axes[iDim - 1];
}

std::vector<axis> hyperCube::getAxes() {
	assert(axesInitialized);
	return axes;
}

void hyperCube::writeAxes(string tag) {
	assert(axesInitialized);
	for (int i = 1; i <= axes.size(); i++) {
		int loc = i - 1;
		int n = axes[loc].n;
		float o = axes[loc].o, d = axes[loc].d;
		string label = axes[loc].label;
		sep_put_data_axis_par(tag.c_str(), &i, &n, &o, &d, label.c_str());
	}
}
void hyperCube::readAxes(string tag) {
	axes.clear();
	int ndim = 0;
	sep_get_number_data_axes(tag.c_str(), &ndim);
	for (int i = 1; i <= ndim; i++) {
		int n;
		float o, d;
		char lab[128];
		sep_get_data_axis_par(tag.c_str(), &i, &n, &o, &d, lab);
		axis tmp = axis(n, o, d, lab);
		axes.push_back(tmp);
	}
	naxis = axes.size();
	axesInitialized = true;
}

int hyperCube::getSize() {
	int res = 1;
	for (int i = 0; i < axes.size(); i++) {
		if (axes[i].n != 0)
			res *= axes[i].n;
	}
	return res;
}

void hyperCube::axesInfo() {
	assert(axesInitialized);
	for (int i = 1; i <= naxis; i++) {
		axes[i - 1].axisInfo(i);
	}
}
bool hyperCube::axesExist() {
	return axesInitialized;
}


// Below is the actual implementation for Vector class
MYVector::MYVector(){}

MYVector::MYVector(int n1, int n2, int n3, int n4, int n5, int n6, int n7) {
	std::vector<int> dims = {n1, n2, n3, n4, n5, n6, n7};
	std::vector<axis> axes_tmp;
	for (int i = 0; i < 6; i++) {
		if (dims[i])
			axes_tmp.push_back(axis(dims[i]));
//		else
//			axes_tmp.push_back(axis());
	}
	init(axes_tmp);
}

MYVector::MYVector(axis ax1, axis ax2, axis ax3, axis ax4, axis ax5, axis ax6, axis ax7) {
	std::vector<axis> axes_tmp;
	if (ax1.n != 0) axes_tmp.push_back(ax1);
	if (ax2.n != 0) axes_tmp.push_back(ax2);
	if (ax3.n != 0) axes_tmp.push_back(ax3);
	if (ax4.n != 0) axes_tmp.push_back(ax4);
	if (ax5.n != 0) axes_tmp.push_back(ax5);
	if (ax6.n != 0) axes_tmp.push_back(ax6);
	if (ax7.n != 0) axes_tmp.push_back(ax7);
	init(axes_tmp);

}
MYVector::MYVector(std::vector<axis> axes) {
	init(axes);
}
MYVector::~MYVector(){
	if (dataPtr == NULL) {
		delete[] dataPtr;
	}
}
void MYVector::allocateMemory() {
	init(getAxes());
}

void MYVector::init(std::vector<axis> axes_tmp) {
//	while (!((axes_tmp.back())->initialized)) {
//		axes_tmp.pop_back();
//	}
	setAxes(axes_tmp);
	ndim = axes_tmp.size();
	nsize = 1;
	for (int i = 0; i < ndim; i++) {
		nvec.push_back(axes_tmp[i].n);
		ovec.push_back(axes_tmp[i].o);
		dvec.push_back(axes_tmp[i].d);
		unitvec.push_back(nsize);
		nsize *= nvec[i];

	}
	std::cout << "nsize = " << nsize << std::endl;
	if (dataPtr == NULL) delete[] dataPtr;
	dataPtr = new float[nsize];
}

void MYVector::readFromFile(string tag) {
	 if (!axesExist()) {
		 readAxes(tag);
	 }
	 sreed(tag.c_str(), dataPtr, sizeof(float) * nsize);
}
void MYVector::writeToFile(string tag) {
	 writeAxes(tag);
	 srite(tag.c_str(), dataPtr, sizeof(float) * nsize);
}

float& MYVector::operator()(int i1, int i2, int i3, int i4, int i5, int i6, int i7) {
	std::vector<int> coord = {i1, i2, i3, i4, i5, i6, i7};
	int loc = 0;
	for (int i = 0; i < ndim; i++) {
		loc += coord[i] * unitvec[i];
	}
	return dataPtr[loc];
}

float& MYVector::operator[](int i)  {
	 return dataPtr[i];
}

void MYVector::clear() {
	for (int i = 0; i < nsize; i++) dataPtr[i] = 0;
	memset(dataPtr, 0, nsize * sizeof(float));
}

// Implementation of linear algebra operator
void clear(MYVector& vec) {
  vec.clear();
}
double innerProd(MYVector& x, MYVector& y) {
	double res = 0;
	float *xPtr = x.getData(), *yPtr = y.getData();
	for (int i = 0; i < x.size(); i++) {
		res += xPtr[i] * yPtr[i];
	}
	return res;
}

void addInv(MYVector& x) {
	float *xPtr = x.getData();
	for (int i = 0; i < x.size(); i++) {
		xPtr[i] = -xPtr[i];
	}
}
// y = a * x + b * y
void linearCombo(float a, const MYVector& x, float b, MYVector& y) {
	float *xPtr = x.getData(), *yPtr = y.getData();
	for (int i = 0; i < x.size(); i++) {
		yPtr[i] = b * yPtr[i] + a * xPtr[i];
	}
}


}
#endif

//	// void random
//	void random(MYVector& X) {
//		float *xptr = X.getData();
//		for (int i = 0; i < X.size(); i++) {
//			xptr[i] = 1.; // TBD
//		}
//	}

//	void set(int index, float val) {
//		dataPtr[index] = val;
//	}

//	// data reference

//	// Standard operation
//	int size() const { return nsize;}
//	std::vector<int> sizes() const {return nvec;};
//	float* getData() const { return NULL;}
//	float& operator[] (int index);
//	float& at(int index) {return dataPtr[index];}
//	void clear() {memset(dataPtr, 0, nsize * sizeof(float));}


//// Linear Algebra part
//// InnerProd
//double innerProd(MYVector<F>& X, MYVector<F>& Y) {
//	double res = 0;
//	F *xptr = X.getData();
//	F *yptr = Y.getData();
//	for (int i = 0; i < X.size(); i++) {
//		res += xptr[i] * yptr[i];
//	}
//	return res;
//}
//
//// inverse
//void addInv(MYVector<F>& X) {
//	F *xptr = X.getData();
//	for (int i = 0; i < X.size(); i++) {
//		xptr[i] *= -1;
//	}
//}
////
//// scale
//void scale(F *a, MYVector<F>& X) {
//	F *xptr = X.getData();
//	for (int i = 0; i < X.size(); i++) {
//		xptr[i] *= a;
//	}
//}


// linear algebra opeartion

//	virtual void linearCombo(float a, numCube  * x, float b, numCube * y) const;
//	virtual double innerProd(numCube * x, numCube * y) const;
//	virtual void addInv(numCube * x) const;
//	virtual void addInv(numCube  * x, dataContainer * xNeg) const ;
//	virtual void scale(float a, numCube * x)const ;
//	virtual void scale(float a, numCube * x,dataContainer * ax)const ;
//	virtual void zeroElement(numCube * x) const ;
//	virtual void random(numCube * x) const ;
//	virtual ~numCube(){ }

