#ifndef NCG_H
#define NCG_H
#include <cmath>
#include <iostream>
#include <cstdint>
#include <cstring>
#include "SEPMYinc.h"
#include "Vector.h"

/*
  NCG

  Nonlinear solver. Including linear CG.
  Use carefully and have fun.

  First Created: 2015-11-04, Yinbin Ma
  Last Modified: 2017-03-15

*/


namespace SEPMY {
int VSIZE;
double NCGMax;
double NCG_inner(float const *d1, float const *d2, int const n);

class Wrapper {
public:
    Wrapper() {}
    virtual ~Wrapper(){}
    virtual void Evaluate(float* model, double& objf, float* grad) {}
    virtual void fOper(bool add, float * model, float *data){}
    virtual void bOper(bool add, float * model, float *data){}
};

class WrapperRosenbrock : public Wrapper {
public:
	WrapperRosenbrock() {}
	void Evaluate(float * model, double& objf, float* grad) {
		// a = 1, b = 100
		float x = model[0];
		float y = model[1];
		objf = (1 - x) * (1 - x) + 100 * (y - x * x) * (y - x * x);
		grad[0] = 2 * (x - 1) + 400 * (x * x - y) * x;
		grad[1] = 200 * (y - x * x);
	}
};

void bound(float *model){}

class Zoom {
public:
    Zoom(int modelSize) {
        this->modelSize = modelSize;
        modelTry = new float[modelSize];
        gradTry = new float[modelSize];
    }
    int zoom(double& alpha, float * model, double& objf, float * grad, int& nsteps,
             float const * model0, float const * dSearch, double const objf0, double const gradAlpha0,
             double alphaLo, double objfLo, double gradAlphaLo,
             double alphaHi, double objfHi, double gradAlphaHi,
             double c1, double c2, Wrapper * Wrap);
private:
    int modelSize;
    void printTry() {}
    double interPolateAlpha(double alphaLo, double objfLo, double gradAlphaLo, double alphaHi, double objfHi, double gradAlphaHi);
    float *modelTry, *gradTry;
};

class LineSearch {
public:
    LineSearch (int modelSize);
    ~LineSearch () {}
    void setAlphaMax(double alphaMax) { this->alphaMax = alphaMax;}
    int lineSearch(double& alpha, float * model, double& objf, float * grad, int& nsteps,
    		float const * model0, float const * dSearch, double const objf0, float const * grad0,
    		double const c1, double const c2, Wrapper * Wrap);

private:
    int modelSize;
    double alphaMax;
    Zoom * ZoomFun;
    int lineNiter;
    float *modelNew, *gradNew;
};

class NCG {
public:
	NCG(int modelSize, Wrapper * Wrap);
	~NCG() {}
	void setPar(int niter, double alphaMax);
	int run(float *model, double * objfs, string tag = "", hyperCube *cube = NULL, float *models = NULL, float * grads = NULL, int * nProp = NULL);

private:
	// parameter that matters
	int modelSize;
	LineSearch * Line;
	Wrapper * Wrap;
	int niter;
	double c1, c2;
	double alpha;

	// function that matters
	void steepestDescent(float * model, double& objf, float * grad, float * dSearch);
	void saveResult(float * model, double objf, float * grad, int iter, int nsteps, double *objfs, float *models, float *grads, int *nProp, bool verb = true);
	void getNewDirection(float * grad, float * gradOld, float *dSearch);
};

// linear conjugate gradient
class LCG {
public:
	LCG(int modelSize, int dataSize, Wrapper * Wrap) {
		this->modelSize = modelSize;
		this->dataSize = dataSize;
		this->Wrap = Wrap;

	}
	~LCG() {}
	void setPar(int niter) {
		this->niter = niter;
	}
	int run(float *model, float * data, double * objfs, float *models = NULL, float * grads = NULL, int * nProp = NULL) {
		// initialization
		memset(model, 0, sizeof(float) * modelSize);
		float * s = new float[dataSize];
		for (int i = 0; i < dataSize; i++) s[i] = -data[i];
		float * Gp = new float[dataSize];
		float * p = new float[modelSize]; memset(p, 0, sizeof(float) * modelSize);
		float * r = new float[modelSize]; Wrap->bOper(false, r, s);
		float * grad = new float[modelSize];
		double objf = 0, beta = 0, alpha, rnormOld = 0, rnormNew = 0;
		rnormOld = NCG_inner(r, r, modelSize);

		// iteration
		for (int iter = 0; iter < niter; iter++) {
			std::cout << "at iter " << iter << ", data error = ";
			std::cout << sqrt(NCG_inner(s, s, modelSize));
			std::cout << ", regularization error ";
			std::cout << sqrt(NCG_inner(&s[modelSize], &s[modelSize], 3 * modelSize));
			std::cout << ", model norm = " << NCG_inner(&s[4 * modelSize], &s[4 * modelSize], modelSize)<< std::endl;
			objfs[iter] = rnormOld;
			// 1. p = -r + beta * p
			for (int i = 0; i < modelSize; i++) {
				p[i] = -r[i] + beta * p[i];
			}
			// 2. alpha
			Wrap->fOper(false, p, Gp);
			alpha =  rnormOld / NCG_inner(Gp, Gp, dataSize);
			// 3. update model
			for (int i = 0; i < modelSize; i++) {
				model[i] = model[i] + alpha * p[i];
			}
			// 4. update s
			for (int i = 0; i < dataSize; i++) {
				s[i] = s[i] + alpha * Gp[i];
			}
			// 5. update r
			Wrap->bOper(false, r, s);
			// 6. update beta
			rnormNew = NCG_inner(r, r, modelSize);
			beta = rnormNew / rnormOld;
			rnormOld = rnormNew;
		}
		return 0;
	}

private:
	// parameter that matters
	int modelSize, dataSize;
	LineSearch * Line;
	Wrapper * Wrap;
	int niter;
	float weight;
};


// below is the actually implementation
int Zoom::zoom(double& alpha, float * model, double& objf, float * grad, int& nsteps,
          float const * model0, float const * dSearch, double const objf0, double const gradAlpha0,
          double alphaLo, double objfLo, double gradAlphaLo,
          double alphaHi, double objfHi, double gradAlphaHi,
          double c1, double c2, Wrapper * Wrap) {
     bool fail = false;
     double alphaTry, objfTry, gradAlphaTry;

	 int nsmax = 4;
     while (fabs(alphaLo - alphaHi) > 1.e-4 && nsteps < nsmax) {
         // bisect
         //double alphaTry = (alphaLo + alphaHi) / 2.;
         std::cout << "Zoom function, ";
     	 alphaTry = interPolateAlpha(alphaLo, objfLo, gradAlphaLo, alphaHi, objfHi, gradAlphaHi);
         // evaluate phi(alpha_j)
         for (int i = 0; i < modelSize; i++) {
             modelTry[i] = alphaTry * dSearch[i] + model0[i];
         }

         Wrap->Evaluate(modelTry, objfTry, gradTry);
         nsteps++;
 		 std::cout << ": alpha = " << alphaTry << ", f/f0 = " << objfTry / NCGMax;
         gradAlphaTry = NCG_inner(gradTry, dSearch, modelSize);
         // branch
         if (objfTry > objf0 + c1 * alphaTry * gradAlpha0 || objfTry >= objfLo || nsteps < nsmax - 1) {
             alphaHi = alphaTry;
             objfHi = objfTry;
             gradAlphaHi = gradAlphaTry;
         } else {
             if (nsteps == nsmax - 1 || fabs(gradAlphaTry) < -c2 * gradAlpha0) {
                 alpha = alphaTry;
                 memcpy(model, modelTry, sizeof(float) * modelSize); // new model
                 objf = objfTry;
                 memcpy(grad, gradTry, sizeof(float) * modelSize);
                 std::cout << ", zoom successful. \n" << std::endl;
                 return 0;
             }
             if (gradAlphaTry * (alphaHi - alphaLo) >= 0) {
                 alphaHi = alphaLo;
                 objfHi = objfLo;
                 gradAlphaHi = gradAlphaLo;
             }
             alphaLo = alphaTry;
             objfLo = objfTry;
             gradAlphaLo = gradAlphaTry;
         }
         std::cout << ", low = " << alphaLo << ", high = " << alphaHi << "\n" << std::endl;
     }

     // failed if arrive here
     alpha = alphaTry;
     memcpy(model, modelTry, sizeof(float) * modelSize); // new model
     objf = objfTry;
     memcpy(grad, gradTry, sizeof(float) * modelSize);
     std::cout << "Zoom failed, f= "<< objf/NCGMax << "\n" << std::endl;
     return -1;
 }


double Zoom::interPolateAlpha(double alphaLo, double objfLo, double gradAlphaLo, double alphaHi, double objfHi, double gradAlphaHi) {
	double bisect = (alphaLo + alphaHi) / 2.;
	double eta = gradAlphaLo + gradAlphaHi - 3 * (objfLo - objfHi) / (alphaLo - alphaHi);
	double sgn = 1.;
	if (alphaHi < alphaLo) sgn = -1.;
	double gamma = sgn * sqrt(eta * eta - gradAlphaLo * gradAlphaHi);
	double interp = alphaHi - (alphaHi - alphaLo) * (gradAlphaHi + gamma - eta) / (gradAlphaHi - gradAlphaLo + 2 * gamma);
	int in_int = 0;
	if (std::isfinite(interp)) {
		if (alphaLo <= bisect) {
			if (interp >= alphaLo && interp <= bisect){
				in_int = 1;
			}
		} else {
			if (interp >= bisect && interp <= alphaLo) {
				in_int = 1;
			}
		}
	}
	if (in_int) {
		std::cout << "cubic  interp";
		return interp;
	} else {
		std::cout << "bisect interp";
		return bisect;
	}
}

LineSearch::LineSearch (int modelSize) {
    this->modelSize = modelSize;
    ZoomFun = new Zoom(modelSize);
    lineNiter = 100;
    modelNew = new float[modelSize];
    gradNew = new float[modelSize];

}

int LineSearch::lineSearch(double& alpha, float * model, double& objf, float * grad, int& nsteps,
 		float const * model0, float const * dSearch, double const objf0, float const * grad0,
 		double const c1, double const c2, Wrapper * Wrap) {
 	// initilization
 	double gradAlpha0 = NCG_inner(grad0, dSearch, modelSize);
 	//std::cout << "    gradAlpha0 = " << gradAlpha0 << std::endl;
 	double alphaOld = 0;
 	double objfOld = objf0;
 	double gradAlphaOld = gradAlpha0;
 	double alphaNew = alphaMax;
    double objfNew, gradAlphaNew;


 	// iteratively minimize the objective function
 	for (int iter = 0; iter < lineNiter; iter++) {
 		std::cout << "Line Search iteration " << iter;
 		// new trial point, evaluate
 		for (int i = 0; i < modelSize; i++) {
 			modelNew[i] = model0[i] + alphaNew * dSearch[i];
 		}
 		Wrap->Evaluate(modelNew, objfNew, gradNew);
 		gradAlphaNew = NCG_inner(gradNew, dSearch, modelSize);
 		nsteps++;
 		std::cout << ": alpha = " << alphaNew << ", f/f0 = " << objfNew / NCGMax;

 		if (objfNew > objf0 + c1 * alphaNew * gradAlpha0 || (objfNew >= objfOld && iter > 0)) {
 			std::cout << ", regular zoom." << std::endl;
 			int flag = ZoomFun->zoom(alpha, model, objf, grad,nsteps,
 					model0, dSearch, objf0, gradAlpha0,
 					alphaOld, objfOld, gradAlphaOld,
 					alphaNew, objfNew, gradAlphaNew,
 					c1, c2, Wrap);
 			return flag;
 		}
 		if (fabs(gradAlphaNew) <= -c2 * gradAlpha0) {
 			std::cout << ", accept trial step.\n" << std::endl;
 			alpha = alphaNew;
 			memcpy(model, modelNew, sizeof(float) * modelSize);
 			objf = objfNew;
 			memcpy(grad, gradNew, sizeof(float) * modelSize);
 			return 0;
 		}
 		if (gradAlphaNew >= 0) {
 		    // reverse order zoom
 			std::cout << ", reverse zoom.\n" << std::endl;
 			int flag = ZoomFun->zoom(alpha, model, objf, grad,nsteps,
 					model0, dSearch, objf0, gradAlpha0,
 					alphaNew, objfNew, gradAlphaNew,
 					alphaOld, objfOld, gradAlphaOld,
 					c1, c2, Wrap);
 			return flag;
 		}
 		// non of above, Expand
 		std::cout << ", expand search bracket.\n" << std::endl;
 		alphaOld = alphaNew;
 		objfOld = objfNew;
 		gradAlphaOld = gradAlphaNew;
 		alphaNew = 2 * alphaNew;
 	}
 	// failed
 		std::cout << "    Line Search Failed.\n" << std::endl;
		alpha = alphaNew;
		memcpy(model, modelNew, sizeof(float) * modelSize);
		objf = objfNew;
		memcpy(grad, gradNew, sizeof(float) * modelSize);
 	return 0;
 }


NCG::NCG(int modelSize, Wrapper * Wrap) {
	this->modelSize = modelSize;
	this->Wrap = Wrap;
	c1 = 0.01;
	c2 = 0.8;
	niter = 50;
	alpha = -1.;
	Line = new LineSearch(modelSize);
	std::cout.precision(3);
	std::cout << std::scientific;
}

void NCG::setPar(int niter, double alpha) {
	this->niter = niter;
	this->alpha = alpha;
	Line->setAlphaMax(alpha);
}

int NCG::run(float *model, double * objfs,  string tag, hyperCube *cube, float *models, float * grads, int * nProp) {
	// initialize necessary
	// parameters that does not matter
	float * dSearch = new float[modelSize];   memset(dSearch, 0, sizeof(float) * modelSize);
	float *grad = new float[modelSize];      memset(grad, 0, sizeof(float) * modelSize);
	float *modelOld = new float[modelSize];
	float *gradOld = new float[modelSize];
	double objfOld = 0, objf = 0;
	int nsteps = 0;


	// First step, evaluate objective function, steepest descent
	steepestDescent(model, objf, grad, dSearch); // compute from model
	saveResult(model, objf, grad, 0, nsteps, objfs, models, grads, nProp, false);
	NCGMax = objf;
	double gnorm2 = NCG_inner(grad, grad, modelSize);
	if (alpha < 0) {
		alpha = 2 * NCGMax / gnorm2;
		Line->setAlphaMax(alpha);
	}
    std::cout << "NCG, starting from f = " << objf << ", gradient norm = " << sqrt(gnorm2) << ", Max iter = " << niter << std::endl;
    // iteratively minimize objective function
	for (int iter = 1; iter <= niter; iter++) {
		nsteps = 0;
		std::cout << "-----------------------------------------------------------------------------------------------------\n" << std::endl;
		// restart if not a descent direction
		if (NCG_inner(grad, dSearch, modelSize) >= 0) {
			steepestDescent(model, objf, grad, dSearch);
			alpha = 0.01;
		}
		// Line search
		memcpy(modelOld, model, sizeof(float) * modelSize);
		objfOld = objf;
		memcpy(gradOld, grad, sizeof(float) * modelSize);
		int flag = Line->lineSearch(alpha, model, objf, grad, nsteps,
				modelOld, dSearch, objfOld, gradOld,
				c1, c2, Wrap); // rework

		Line->setAlphaMax(alpha);
		// save and print line search result
        if (tag != "") {
            srite(tag.c_str(), model, sizeof(float) * modelSize);
            cube->writeAxes(tag.c_str());
        }
		saveResult(model, objf, grad, iter, nsteps, objfs, models, grads, nProp);
		// get new search direction
		getNewDirection(grad, gradOld, dSearch);
		if (flag < 0) {
			memset(dSearch, 0, sizeof(float) * modelSize);
			alpha = objf / NCG_inner(grad, grad, modelSize);
			alpha = 0.01;
			Line->setAlphaMax(alpha);
		}
	}
	return 0;
}


void NCG::steepestDescent(float * model, double& objf, float * grad, float * dSearch) {
	std::cout << "NCG: start from steepest descent" << std::endl;
	Wrap->Evaluate(model, objf, grad);
	for (int i = 0; i < modelSize; i++) dSearch[i] = -grad[i];
}
void NCG::saveResult(float * model, double objf, float * grad, int iter, int nsteps, double *objfs, float *models, float *grads, int *nProp, bool verb) {
	objfs[iter] = objf;
	int low = 0;
	if (nProp) nProp[iter] = nProp[std::max(iter, low)] + nsteps;
	if (models) memcpy(&models[iter * modelSize], model, sizeof(float) * modelSize);
	if (grads) memcpy(&grads[iter * modelSize], grad, sizeof(float) * modelSize);
	if (verb) {
		std::cout << "Summary of NCG iteration " << iter << ": alpha = " << alpha << ", f/f0 = " << objf / NCGMax << ", nsteps = " << nsteps;
	}

}
void NCG::getNewDirection(float * grad, float * gradOld, float *dSearch) {
	float *y = new float[modelSize];
	for (int i = 0; i < modelSize; i++) {
		y[i] = grad[i] - gradOld[i];
	}
	double beta;
	double nmgprevsq = NCG_inner(gradOld, gradOld, modelSize);
	double beta_pr = NCG_inner(grad, y, modelSize) / nmgprevsq;
	double beta_fr = NCG_inner(grad, grad, modelSize) / nmgprevsq;
	if (beta_pr < -beta_fr) {
		beta = -beta_fr;
	} else if (beta_pr > beta_fr) {
		beta = beta_fr;
	} else {
		beta = beta_pr;
	}
	for (int i = 0; i < modelSize; i++) {
		dSearch[i] = beta * dSearch[i] - grad[i];
	}
	std::cout << ",  beta  = " << beta <<std::endl;
	delete[] y;
}

double NCG_inner(float const *d1, float const *d2, int const n){
    double result = 0.;
    for(int i = 0; i < n; i++)
        result += d1[i] * d2[i];
    return result;
};

}

#endif // NCG_H
