#ifndef WAVEPROP3D_H
#define WAVEPROP3D_H 1
#include "stdio.h"
#include "math.h"
#include <unordered_map>
#include "waveCore.h"

namespace SEPMY {

class prop3DVp {
public:
	// dt = [ms], dx = [m], v = [m/ms]
	prop3DVp();
	prop3DVp(int nt, int nx, int ny, int nz, float dt, float dx, float dy = -1.0, float dz = -1.0, float ot = 0.0, float ox = 0.0, float oy = 0.0, float oz = 0.0);
	~prop3DVp();
	void setModel(float *vel);
	void getInfo();
	void setSource(float *source, float *sx, float *sy, float *sz, int ns = 1);  //source[ns][nt]
	void setData(float *shot, float *sx, float *sy, float *sz, int nd = 1);      // data[nr][nt]
	void sourceProp(float *wave = NULL, float *taper = NULL, int recordEvery = 1);
	void sourceReverse(float *wave = NULL, float *taper = NULL, int recordEvery = 1);

private:
	int nz, ny, nx, nt;
	float dz, dy, dx, dt, ox, oy, oz, ot;
	float *vel = NULL, *velDt2 = NULL, *velDtDx = NULL;
	int nsShotMax, shotCount;
	float *shotBin;
	int *shotGeo;
	float *prev, *curr, *next, *p0, *p1;
};


// Below is the actual implementation

// construct, initialization
prop3DVp::prop3DVp() {

}
prop3DVp::prop3DVp(int nt, int nx, int ny, int nz, float dt, float dx, float dy, float dz, float ot, float ox, float oy, float oz) {
	this->nt = nt;
	this->nx = nx;
	this->ny = ny;
	this->nz = nz;
	this->dt = dt;
	this->dx = dx;
	this->dy = (dy > 0 ? dy : dx);
	this->dz = (dz > 0 ? dz : dx);
	this->ox = ox;
	this->oy = oy;
	this->oz = oz;
	this->ot = ot;
	vel = NULL;
	velDt2 = NULL;
	velDtDx = NULL;
	nsShotMax = 0;
	shotBin = NULL;
	shotGeo = NULL;
	prev = new float[nx * ny * nz];
	curr = new float[nx * ny * nz];
	next = new float[nx * ny * nz];
	p0 = new float[nx * ny * nz];
	p1 = new float[nx * ny * nz];
	shotCount = 0;

}

// set up velocity model
void prop3DVp::setModel(float *_vel) {

	if (vel == NULL) vel = new float[nx * ny * nz];
	if (velDt2 == NULL) velDt2 = new float[nx * ny * nz];
	if (velDtDx == NULL) velDtDx = new float[nx * ny * nz];
	for (int i = 0; i < nz * nx * ny; i++) {
		double v = _vel[i];
		vel[i] = v;
		velDt2[i] = v * v * dt * dt;
		velDtDx[i] = v * v * dt * dt / (dx * dx);
	}

}

void prop3DVp::getInfo() {
	std::cout << "\n Printing Propogation parameters" << std::endl;
	std::cout << "nt = " << nt << ", dt = " << dt << ", ot = " << ot << std::endl;
	std::cout << "nz = " << nz << ", dz = " << dz << ", oz = " << oz << std::endl;
	std::cout << "ny = " << ny << ", dy = " << dy << ", oy = " << oy << std::endl;
	std::cout << "nx = " << nx << ", dx = " << dx << ", ox = " << ox << std::endl;

}

// set up source wavelet
void prop3DVp::setSource(float *source, float *sxPtr, float *syPtr, float *szPtr, int ns) {
	// expand if out of allocated moeory
	if (ns > nsShotMax) {
		nsShotMax = ns;
		if (shotBin) delete[] shotBin;
		if (shotGeo) delete[] shotGeo;
		shotBin = new float[nsShotMax * nt * 8];
		shotGeo = new int[nsShotMax  * 8];
	}

	// now assignment
	memset(shotBin, 0, sizeof(float) * nsShotMax * nt * 8);
	memset(shotGeo, 0, sizeof(int) * nsShotMax * 8);
	unordered_map<int, int> maps;
	int count = 0;
	for (int is = 0; is < ns; is++) {
		float sx = sxPtr[is], sy = syPtr[is], sz = szPtr[is];
		int xloc = (int) ((sx - ox) / dx);
		int yloc = (int) ((sy - oy) / dy);
		int zloc = (int) ((sz - oz) / dz);
		float wx = 1. - (sx - ox - xloc * dx) / dx; // wx [0,1], 1: close to left, 0: close to right
		float wy = 1. - (sy - oy - yloc * dy) / dy;
		float wz = 1. - (sz - oz - zloc * dz) / dz;

		int locs[8];
		locs[0] =       zloc * nx * ny +       yloc * nx + xloc;
		locs[1] =       zloc * nx * ny +       yloc * nx + xloc + 1;
		locs[2] =       zloc * nx * ny + (yloc + 1) * nx + xloc;
		locs[3] =       zloc * nx * ny + (yloc + 1) * nx + xloc + 1;
		locs[4] = (zloc + 1) * nx * ny +       yloc * nx + xloc;
		locs[5] = (zloc + 1) * nx * ny +       yloc * nx + xloc + 1;
		locs[6] = (zloc + 1) * nx * ny + (yloc + 1) * nx + xloc;
		locs[7] = (zloc + 1) * nx * ny + (yloc + 1) * nx + xloc + 1;
		float weights[8];
		weights[0] = wz * wy * wx;
		weights[1] = wz * wy * (1 - wx);
		weights[2] = wz * (1 - wy) * wx;
		weights[3] = wz * (1 - wy) * (1 - wx);
		weights[4] = (1 - wz) * wy * wx;
		weights[5] = (1 - wz) * wy * (1 - wx);
		weights[6] = (1 - wz) * (1 - wy) * wx;
		weights[7] = (1 - wz) * (1 - wy) * (1 - wx);
		for (int i = 0; i < 8; i++) {
			if (weights[i] < 0.0000001) continue;
			if (!maps.count(locs[i])) maps[locs[i]] = count++;
			int locT = maps[locs[i]];
			shotGeo[locT] = locs[i];
			for (int it = 0; it < nt; it++) {
				shotBin[locT * nt + it] += weights[i] * source[is * nt + it];
			}
		}
		shotCount = count;
	}

}
void prop3DVp::setData(float *shot, float *sx, float *sy, float *sz, int nd) {

}
void prop3DVp::sourceProp(float *wave, float *taper, int recordEvery) {

	bool saveFlag = (wave == NULL ? false : true);
	bool taperFlag = (taper == NULL ? false : true);

	if (saveFlag) {
		memset(wave, 0, sizeof(float) * nz * ny * nx);  // t0
		cout << "start wave prop, we want to save the wavefield" << endl;
	} else {
		cout << "fuck you" << endl;
	}
	memset(prev, 0, sizeof(float) * nz * nx * ny);
	memset(curr, 0, sizeof(float) * nz * nx * ny);
	memset(next, 0, sizeof(float) * nz * nx * ny);

	// time marching, compute at wave[it]
	for (int it = 1; it <= nt; it++) {
		cout << "it = " << it << ", out of " << nt << endl;
		int id = it / recordEvery;
		int ii = it % recordEvery;
		marchingKernel3D(prev, curr, next, velDtDx, nx, ny, nz); // time marching
		inject(next, shotBin, shotGeo, shotCount, it - 1, nt);           // inject source wavelet
		if (taperFlag) {
			for (int i = 0; i < nz * nx * ny; i++) {
				next[i] *= taper[i];
				curr[i] *= taper[i];
			}
		}
		if (saveFlag && ii == 0 && it < nt) {
			memcpy(&wave[id * nz * ny * nx], next, sizeof(float) * nz * ny * nx);
		}
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
	}
	memcpy(p0, prev, sizeof(float) * nx * ny * nz);
	memcpy(p1, curr, sizeof(float) * nx * ny * nz);
}

void prop3DVp::sourceReverse(float *wave, float *taper, int recordEvery) {

	bool saveFlag = (wave == NULL ? false : true);
	bool taperFlag = (taper == NULL ? false : true);

	if (saveFlag) {
		memset(wave, 0, sizeof(float) * nz * ny * nx);  // t0
		cout << "start wave prop, we want to save the wavefield" << endl;
	} else {
		cout << "fuck you" << endl;
	}
	memset(prev, 0, sizeof(float) * nz * nx * ny);
	memset(curr, 0, sizeof(float) * nz * nx * ny);
	memset(next, 0, sizeof(float) * nz * nx * ny);
	memcpy(prev, p1, sizeof(float) * nx * ny * nz);
	memcpy(curr, p0, sizeof(float) * nx * ny * nz);
	// time marching, compute at wave[it]
	for (int it = nt - 2; it >= 0; it--) {

		int id = (nt - it - 1) / recordEvery;
		int ii = it % recordEvery;
		marchingKernel3D(prev, curr, next, velDtDx, nx, ny, nz);        // time marching
		inject(next, shotBin, shotGeo, shotCount, it + 1, nt);           // inject source wavelet
		if (taperFlag) {
			for (int i = 0; i < nz * nx * ny; i++) {
				next[i] *= taper[i];
				curr[i] *= taper[i];
			}
		}
		if (saveFlag && ii == 0 && it < nt) {
			memcpy(&wave[id * nz * ny * nx], next, sizeof(float) * nz * ny * nx);
		}
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
	}
}

}
#endif
