#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "SEPMYinc.h"
#include "sincCoef.h"
#include "Vector.h"
extern "C" {
#include "waveKernelVTI2d.cu"
}
/*
  Wave Prop2dVTIGPUagation module.

  The most important module.
  Designed for single shot propagation.
  Use carefully and have fun.

  First Created: 2015-11-04, Yinbin Ma
  Last Modified: 2018-01-23

*/
using namespace std;

namespace SEPMY {
class prop2D {
public:
	// dt = [ms], dx =[m], v = [m/ms] = [km/s]
	prop2D(int nt, int nx, int nz, float dt, float dx, float dz = -1.0, float ot = 0.0, float ox = 0.0, float oz = 0.0, float dtProp = 1.0, int maxNumSource = 1, int maxNumReceiver = 1, int nhx = 1);
	~prop2D();
	void getInfo();
	void resetAcquisition(int maxNumSource, int maxNumReceiver);
	void setModel(float *vel, float *epsilon, float *delta);
	void setData(float *data, float *geom, int nrec, float oxIn = 0.0);
	void extractData(float *data, float *geom, int nrec, float oxIn = 0.0);
	void synData(float *data, float *geom, int nrec, float oxIn = 0.0);
	void setShot(float *source, float *sxs, float *szs, int nsource = 1);
	void sourceProp(float *p0, float *p1, int startT, int endT, float *wave = NULL);
	void rtmOper(float *img);
	void gradOper(float *velImg, float *epsImg, float *deltaImg);
private:
	void initPar();
	void allocateMem();
	void setSincTable();
	void transferSymbol();
	void resetTmpArray();
	void setTaper();
	void sourcePropInternal();
	float *prev, *curr, *next, *prevR, *currR, *nextR, *p0, *p1;
	float *vel, *epsilon, *delta;
	float *velDt2, *velDtDx, *epsilonTerm, *deltaTerm;
	float *image, *velGrad, *epsilonGrad, *deltaGrad;
	float *taper;
	float *wavelet;
	int *sourceLoc;
	float *dataShot, *dataWeight;
	int *dataLoc;
	float *dataTmp;
	int *locTmp;
	float *tables;
	int nt, nz, nx, nhx, msize;
	float dt, dz, dx, ds, dr, dtProp;
	float ot, oz, ox, os;
	int tsample, nsinc;
	int maxNumReceiver, maxNumSource;
	dim3 dimGrid, dimBlock, dimGridS, dimBlockS, dimGridR, dimBlockR;
	int startT, endT;
};


prop2D::prop2D(int nt, int nx, int nz, float dt, float dx, float dz, float ot, float ox, float oz, float dtProp, int maxNumSource, int maxNumReceiver, int nhx) {
	this->nt = nt;
	this->nx = nx;
	this->nz = nz;
	msize = nx * nz;
	this->dt = dt;
	this->dx = dx;
	this->dz = (dz < 0 ? dx : dz);
	this->ot = ot;
	this->ox = ox;
	this->oz = oz;
	this->dtProp = dtProp;
	this->maxNumSource   = maxNumSource;
	this->maxNumReceiver = maxNumReceiver;
	this->nhx = nhx;
	initPar();
	allocateMem();
	setSincTable();
	transferSymbol();
	setTaper();
}

// initialize propagation sampling, GPU grid, etc
// source use (N, 1, 1) grid, receiver use (N, 1, N) grid
void prop2D::initPar() {
	tsample = std::round(dt / dtProp);
	cout << "tsample = " << tsample << endl;
	int nblocksZ = (nz - 2 * FAT) / BLOCK_SIZE;
	int nblocksX = (nx - 2 * FAT) / BLOCK_SIZE;
	dimGrid = dim3(nblocksX, 1, nblocksZ);
	dimBlock = dim3(BLOCK_SIZE, 1, BLOCK_SIZE);
	//cout << "nblocksZ = " << nblocksZ << ", nblocksX = " << nblocksX << endl;
	//cout << "BLOCK_SIZE = " << BLOCK_SIZE <<  endl;
	int nblocksS = (maxNumSource * 4 + BLOCK_SIZE - 1) / BLOCK_SIZE;
	dimGridS = dim3(nblocksS, 1, 1);
	dimBlockS = dim3(BLOCK_SIZE, 1, 1);
	float tmp = maxNumReceiver / (1.0 * BLOCK_SIZE);
	int nblocksR = ceil(tmp);
	dimGridR = dim3(nblocksR, 1, 1);
	dimBlockR = dim3(BLOCK_SIZE, 1, 4);
}

// allocate memory on GPU and CPU
void prop2D::allocateMem() {
	// GPU memory for computation
	cudaMalloc((void**) &prev, msize * 2 * sizeof(float));
	cudaMalloc((void**) &curr, msize * 2 * sizeof(float));
	cudaMalloc((void**) &next, msize * 2 * sizeof(float));
	cudaMalloc((void**) &prevR, msize * 2 * sizeof(float));
	cudaMalloc((void**) &currR, msize * 2 * sizeof(float));
	cudaMalloc((void**) &nextR, msize * 2 * sizeof(float));

	cudaMalloc((void**) &velDt2, msize * sizeof(float));
	cudaMalloc((void**) &velDtDx, msize * sizeof(float));
	cudaMalloc((void**) &epsilonTerm, msize * sizeof(float));
	cudaMalloc((void**) &deltaTerm, msize * sizeof(float));

	cudaMalloc((void**) &image, msize * nhx * sizeof(float));
	cudaMalloc((void**) &wavelet, 4 * maxNumSource * (nt * 2 + 1) * sizeof(float));
	cudaMalloc((void**) &sourceLoc, 4 * maxNumSource * sizeof(int));
	cudaMalloc((void**) &dataShot, 4 * nt * maxNumReceiver * sizeof(float));
	cudaMalloc((void**) &dataLoc, 4 * maxNumReceiver * sizeof(int));

	cudaMalloc((void**) &velGrad,     msize * sizeof(float));
	cudaMalloc((void**) &epsilonGrad, msize * sizeof(float));
	cudaMalloc((void**) &deltaGrad,   msize * sizeof(float));

	// CPU memory for IO with main function
	vel = new float[msize];
	epsilon = new float[msize];
	delta = new float[msize];
	dataTmp = new float[(maxNumSource + maxNumReceiver) * 4 * (2 * nt + 1)];
	locTmp = new int[(maxNumSource + maxNumReceiver) * 4];
}

void prop2D::setSincTable() {
	nsinc = 8;
	sincCoef myr(tsample, nsinc);

	float *tmpTables = new float[tsample * nsinc];
	for (int i = 0; i < tsample; i++) {
		//memcpy(&tmpTables[i * nsinc], myr.table[i], sizeof(float) * nsinc);
		for (int j = 0; j < nsinc; j++) {
			tmpTables[i * nsinc + j] = myr.table[i][j];
			//cout << tmpTables[i*nsinc + j]  << ", ";
		}
		//cout << endl;
	}

	cudaMalloc((void**) &tables, tsample * nsinc * sizeof(float));
	cudaMemcpy(tables, tmpTables, tsample * nsinc * sizeof(float), cudaMemcpyHostToDevice);
	delete[] tmpTables;
}

void prop2D::transferSymbol() {
	cudaMemcpyToSymbol(nzGPU, &nz, sizeof(int));
	cudaMemcpyToSymbol(nxGPU, &nx, sizeof(int));
	cudaMemcpyToSymbol(ntGPU, &nt, sizeof(int));

	cudaMemcpyToSymbol(sourceGPU, &wavelet, sizeof(float *));
	cudaMemcpyToSymbol(sourceLocGPU, &sourceLoc, sizeof(int *));


	cudaMemcpyToSymbol(dataShotGPU, &dataShot, sizeof(float *));
	cudaMemcpyToSymbol(dataLocGPU, &dataLoc, sizeof(int *));

	cudaMemcpyToSymbol(velDt2GPU, &velDt2, sizeof(float *));
	cudaMemcpyToSymbol(velDtDxGPU, &velDtDx, sizeof(float *));
	cudaMemcpyToSymbol(epsilonTermGPU, &epsilonTerm, sizeof(float *));
	cudaMemcpyToSymbol(deltaTermGPU, &deltaTerm, sizeof(float *));

	cudaMemcpyToSymbol(nsincGPU, &nsinc, sizeof(int));
	cudaMemcpyToSymbol(sinc_s_table, &tables, sizeof(float *));
	cudaMemcpyToSymbol(nrGPU, &maxNumReceiver, sizeof(int));

    int hx0 = nhx / 2;
    cudaMemcpyToSymbol(hx0GPU, &hx0, sizeof(int));
}

void prop2D::getInfo() {
	std::cout << "\n Printing Propogation parameters" << std::endl;
	std::cout << "nt = " << nt << ", dt = " << dt << ", ot = " << ot << std::endl;
	std::cout << "nz = " << nz << ", dz = " << dz << ", oz = " << oz << std::endl;
	std::cout << "nx = " << nx << ", dx = " << dx << ", ox = " << ox << std::endl;
	std::cout << "maxNumSource = " << maxNumSource << std::endl;
	std::cout << "maxNumReceiver = " << maxNumReceiver << std::endl;
}

void prop2D::resetAcquisition(int maxNumSource, int maxNumReceiver) {
	this->maxNumSource = maxNumSource;
	this->maxNumReceiver = maxNumReceiver;
}

void prop2D::setModel(float *_vel, float *_epsilon, float *_delta) {
	// copy over the model
	memcpy(vel, _vel, sizeof(float) * msize);
	memcpy(epsilon, _epsilon, sizeof(float) * msize);
	memcpy(delta, _delta, sizeof(float) * msize);

	// precompute coefficients
	float *tmpvelDt2  = new float[msize];
	float *tmpvelDtDx = new float[msize];
	float* tmpepsilonTerm = new float[msize];
	float* tmpdeltaTerm   = new float[msize];
	for (int i = 0; i < nz * nx; i++) {
		tmpvelDt2[i] = vel[i] * vel[i] * dtProp * dtProp;
		tmpvelDtDx[i] = tmpvelDt2[i] / (dx * dx);
		tmpepsilonTerm[i] = (1.0 + 2.0 * epsilon[i]) * tmpvelDtDx[i];
		tmpdeltaTerm[i] = sqrt(1.0 + 2.0 * delta[i]) * tmpvelDtDx[i];
	}

	// copy to gpu memory
	cudaMemcpy(velDt2, tmpvelDt2, sizeof(float) * nz * nx, cudaMemcpyHostToDevice);
	cudaMemcpy(velDtDx, tmpvelDtDx, sizeof(float) * nz * nx, cudaMemcpyHostToDevice);
	cudaMemcpy(deltaTerm, tmpdeltaTerm, sizeof(float) * nz * nx, cudaMemcpyHostToDevice);
	cudaMemcpy(epsilonTerm, tmpepsilonTerm, sizeof(float) * nz * nx, cudaMemcpyHostToDevice);

	// clean CPU temperary memory
	delete[] tmpvelDtDx;
	delete[] tmpvelDt2;
	delete[] tmpepsilonTerm;
	delete[] tmpdeltaTerm;
}

// map source to regular grid using linear interpolation
void prop2D::setShot(float *source, float *sxs, float *szs, int nsource) {
	// increase if exceed max capacity
	if (nsource > maxNumSource) {
		resetAcquisition(nsource, maxNumReceiver);
	}

	float maps[msize];
	for (int i = 0; i < msize; i++) {
		maps[i] = -1;
	}
	int count = 0;

	// for each source point, interpolating to 4 neighbours, push data to the neigh
	memset(dataTmp, 0, sizeof(float) * maxNumSource * 4 * (nt * 2 + 1));
	memset(locTmp, 0, sizeof(int) * 4 * maxNumSource);
	for (int is = 0; is < nsource; is++) {
		float sx = sxs[is];
		float sz = szs[is];
		int xloc = (int) ((sx - ox) / dx);
		int zloc = (int) ((sz - oz) / dz);
		float wx = 1 - (sx - ox - xloc * dx) / dx;
		float wz = 1 - (sz - oz - zloc * dz) / dz;
		int locs[4];
		locs[0] = zloc * nx + xloc;
		locs[1] = zloc * nx + xloc + 1;
		locs[2] = (zloc + 1) * nx + xloc;
		locs[3] = (zloc + 1) * nx + xloc + 1;
		float weights[4];
		weights[0] = wz * wx;
		weights[1] = wz * (1 - wx);
		weights[2] = (1 - wz) * wx;
		weights[3] = (1 - wz) * (1 - wx);
		for (int i = 0; i < 4; i++) {
			if (weights[i] < 0.00001) continue;
			if (maps[locs[i]] == -1) {
				maps[locs[i]] = count;
				count++;
			}
			int loc = maps[locs[i]];
			locTmp[loc] = locs[i];
			for (int it = 0; it < 2 * nt + 1; it++) {
				dataTmp[loc * (2 * nt + 1) + it] += weights[i] * source[is * (2 * nt + 1) + it] * dtProp * dtProp * 1.5 * 1.5;
			}
		}
	}
	cudaMemcpy(wavelet, dataTmp, sizeof(float) * (2 * nt + 1) * maxNumSource * 4, cudaMemcpyHostToDevice);
	cudaMemcpy(sourceLoc, locTmp, sizeof(int) * maxNumSource * 4, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(countSGPU, &count, sizeof(int));
}

// I really do not like this function
void prop2D::setData(float *data, float *geom, int nrec, float oxIn) {
	float maps[msize];
	for (int i = 0; i < msize; i++) {
		maps[i] = -1;
	}
	int count = 0;
	memset(dataTmp, 0, sizeof(float) * maxNumReceiver * 4 * nt);
	memset(locTmp, 0, sizeof(int) * 4 * maxNumReceiver);
	for (int ir = 0; ir < nrec; ir++) {
		if (geom[ir * 7 + 6] == 0 || geom[ir * 7 + 3] < oxIn) continue;
		float rx = geom[ir * 7 + 3];
		float rz = geom[ir * 7 + 5];
		int xloc = (int) ((rx - ox) / dx);
		int zloc = (int) ((rz - oz) / dz);
		float wx = 1 - (rx - ox - xloc * dx) / dx;
		float wz = 1 - (rz - oz - zloc * dz) / dz;
		int locs[4];
		locs[0] = zloc * nx + xloc;
		locs[1] = zloc * nx + xloc + 1;
		locs[2] = (zloc + 1) * nx + xloc;
		locs[3] = (zloc + 1) * nx + xloc + 1;
		float weights[4];
		weights[0] = wz * wx;
		weights[1] = wz * (1 - wx);
		weights[2] = (1 - wz) * wx;
		weights[3] = (1 - wz) * (1 - wx);
		for (int i = 0; i < 4; i++) {
			if (weights[i] < 0.00001) continue;
			if (maps[locs[i]] == -1) {
				maps[locs[i]] = count;
				count++;
			}
			int loc = maps[locs[i]];
			locTmp[loc] = locs[i];
			for (int it = 0; it < nt; it++) {
				dataTmp[loc * nt + it] += weights[i] * data[ir * nt + it];
			}
		}
	}
	cudaMemcpy(dataShot, dataTmp, sizeof(float) * maxNumReceiver * nt * 4, cudaMemcpyHostToDevice);
	cudaMemcpy(dataLoc, locTmp, sizeof(int) * maxNumReceiver * 4, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(countGPU, &count, sizeof(int));
}

// I really do not like this function
void prop2D::extractData(float *data, float *geom, int nrec, float oxIn) {
	float maps[msize];
	for (int i = 0; i < msize; i++) {
		maps[i] = -1;
	}
	int count = 0;
	memset(dataTmp, 0, sizeof(float) * maxNumReceiver * 4 * nt);
	memset(locTmp, 0, sizeof(int) * 4 * maxNumReceiver);
	cudaMemcpy(dataTmp, dataShot, sizeof(float) * maxNumReceiver * nt * 4, cudaMemcpyDeviceToHost);
	cudaMemcpy(locTmp, dataLoc, sizeof(int) * maxNumReceiver * 4, cudaMemcpyDeviceToHost);
	double energy = 0;
	for (int ir = 0; ir < nrec; ir++) {
		if (geom[ir * 7 + 6] == 0 || geom[ir * 7 + 3] < oxIn) continue;
		float rx = geom[ir * 7 + 3];
		float rz = geom[ir * 7 + 5];
		int xloc = (int) ((rx - ox) / dx);
		int zloc = (int) ((rz - oz) / dz);
		float wx = 1 - (rx - ox - xloc * dx) / dx;
		float wz = 1 - (rz - oz - zloc * dz) / dz;
		int locs[4];
		locs[0] = zloc * nx + xloc;
		locs[1] = zloc * nx + xloc + 1;
		locs[2] = (zloc + 1) * nx + xloc;
		locs[3] = (zloc + 1) * nx + xloc + 1;
		float weights[4];
		weights[0] = wz * wx;
		weights[1] = wz * (1 - wx);
		weights[2] = (1 - wz) * wx;
		weights[3] = (1 - wz) * (1 - wx);
		for (int i = 0; i < 4; i++) {
			if (weights[i] < 0.00001) continue;
			if (maps[locs[i]] == -1) {
				maps[locs[i]] = count;
				count++;
			}
			int loc = maps[locs[i]];
			locTmp[loc] = locs[i];
			for (int it = 0; it < nt; it++) {
				data[ir * nt + it] += dataTmp[loc * nt + it] * weights[i];
			}
		}
	}
	for (int i = 0; i < nrec * nt; i++)
		energy += data[i] * data[i];
	cout << "energy = " << energy << endl;

}

void prop2D::setTaper() {
    float temp[nx * nz];
    int n1 = nx;
    int n2 = nz;
    int nb = 60;
    float alpha = 0.03;
    //std::cout << "n1 = " << n1 << ", n2 = " << n2 << std::endl;
	  for(int i = 0; i < n1 * n2; i++) {
		  temp[i] = 1.0;
	  }

	  // TOP and BOTTOM
	  for (int i2 = 0; i2 < nb; i2++) {
	      for (int i1 = nb; i1 < n1 - nb; i1++) {
	          double taps = 1. * (nb - i2) / (1.0 * nb);
	          taps = (1. - alpha) + alpha * cos(PI * taps);
	          temp[i2 * n1 + i1] = taps;
	          temp[(n2 - i2 - 1) * n1 + i1] = taps;
	      }
	  }
	  // left and right
	  for (int i1 = 0; i1 < nb; i1++) {
	      for (int i2 = nb; i2 < n2 - nb; i2++){
	          double taps = 1. * (nb - i1) / (1.0 * nb);
	          taps = (1. - alpha) + alpha * cos(PI * taps);
	          temp[i2 * n1 + i1] = taps;
	          temp[i2 * n1 + n1 - i1 - 1] = taps;
	      }
	  }
	  // Corner
	  for (int i2 = 0; i2 < nb; i2++) {
	      for (int i1 = 0; i1 < nb; i1++) {
		      double taps = 1.0 * (nb - i1) * (nb - i1) + 1.0 * (nb - i2) * (nb - i2);
		      taps = sqrt(taps) / (1.0 * nb);
		      taps = (1. - alpha) + alpha * cos(PI * taps);
		      temp[i2 * n1 + i1] = taps;
		      temp[i2 * n1 + n1 - i1 - 1] = taps;
		      temp[(n2 - i2 - 1) * n1 + i1] = taps;
		      temp[(n2 - i2 - 1) * n1 + n1 - i1 - 1] = taps;
	      }
	  }
	cudaMalloc((void**) &taper, (nz * nx) * sizeof(float));
	cudaMemcpy(taper, temp, sizeof(float) * nx * nz, cudaMemcpyHostToDevice);
}



// do the source propagation
void prop2D::sourcePropInternal() {
	startT = -200;
	endT = nt - 5;
	resetTmpArray();
	for (int it = startT * tsample; it <= endT * tsample; it++) {
		int id = (it + nt * tsample) / tsample;
		int ii = (it + nt * tsample) % tsample;
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prev, curr, next);
		injectWaveletGPU<<<dimGridS, dimBlockS>>>(id, ii, next);
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
	}
	swap(prev, curr);
}

// do the source propagation
void prop2D::synData(float *data, float *geom, int nrec, float oxIn) {
	startT = -200;
	endT = nt - 5;
	resetTmpArray();
	cudaMemset(dataShot, 0, sizeof(float) * maxNumReceiver * nt * 4);

	for (int it = startT * tsample; it <= endT * tsample; it++) {

		int id = (it + nt * tsample) / tsample;
		int ii = (it + nt * tsample) % tsample;
		//cout << "id = " << id << endl;
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prev, curr, next);
		injectWaveletGPU<<<dimGridS, dimBlockS>>>(id, ii, next);
		if (ii == 0 && it >= 0) {
			extractDataKernelGPU<<<dimGridR, dimBlockR>>>(id, ii, next);
			//cudaMemcpy(data, next, sizeof(float) * msize, cudaMemcpyDeviceToHost);
	    	//srite("wave", data, sizeof(float) * nx * nz);
		}
		taperGPU<<<dimGrid, dimBlock>>>(curr, next, taper);
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
	}
	extractData(data, geom, nrec, oxIn);
}

// update img, not just fill
void prop2D::rtmOper(float *img) {
	cout << "running rtm" << endl;
	startT = 3;
	endT = nt - 5;
	resetTmpArray();
	cudaMemcpy(image, img, sizeof(float) * msize * nhx, cudaMemcpyHostToDevice);
	sourcePropInternal();
	for (int it = endT * tsample; it >= startT * tsample; it--) {
		int id = it / tsample;
		int ii = it % tsample;
		//source
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prev, curr, next);
		injectWaveletGPU<<<dimGridS, dimBlockS>>>(id + nt, ii, next);
		// receiver
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prevR, currR, nextR);
		injectDataKernelGPU<<<dimGridR, dimBlockR>>>(id, ii, nextR);
		taperGPU<<<dimGrid, dimBlock>>>(currR, nextR, taper);
		// imaging condition
		if (ii == 0 && it > 0 && id < nt) {
			for (int ihx = 0; ihx < nhx; ihx++) {
				img_hx_kernel<<<dimGrid, dimBlock>>>(image, prev, curr, next, currR, ihx - nhx / 2);
			}
		}
		// rotate pointer
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
		tmp = prevR;
		prevR = currR;
		currR = nextR;
		nextR = tmp;
	}
	cudaMemcpy(img, image, sizeof(float) * msize * nhx, cudaMemcpyDeviceToHost);
	//cudaMemcpy(img, nextR, sizeof(float) * msize * nhx, cudaMemcpyDeviceToHost);
}

void prop2D::gradOper(float *velImg, float *epsilonImg, float *deltaImg) {
	cout << "running gradient" << endl;
	startT = 3;
	endT = nt - 5;
	resetTmpArray();
	cudaMemcpy(velGrad, velImg, sizeof(float) * msize, cudaMemcpyHostToDevice);
	cudaMemcpy(epsilonGrad, epsilonImg, sizeof(float) * msize, cudaMemcpyHostToDevice);
	cudaMemcpy(deltaGrad, deltaImg, sizeof(float) * msize, cudaMemcpyHostToDevice);

	sourcePropInternal();
	for (int it = endT * tsample; it >= startT * tsample; it--) {
		int id = it / tsample;
		int ii = it % tsample;
		//source
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prev, curr, next);
		injectWaveletGPU<<<dimGridS, dimBlockS>>>(id + nt, ii, next);
		// receiver
		marchingKernelGPU<<<dimGrid, dimBlock>>>(prevR, currR, nextR);
		injectDataKernelGPU<<<dimGridR, dimBlockR>>>(id, ii, nextR);
		taperGPU<<<dimGrid, dimBlock>>>(currR, nextR, taper);
		// imaging condition
		if (ii == 0 && it > 0 && id < nt) {
			grad_kernel<<<dimGrid, dimBlock>>>(velGrad, epsilonGrad, deltaGrad, prev, curr, next, currR);
		}
		// rotate pointer
		float *tmp = prev;
		prev = curr;
		curr = next;
		next = tmp;
		tmp = prevR;
		prevR = currR;
		currR = nextR;
		nextR = tmp;
	}
	cudaMemcpy(velImg,     velGrad,     sizeof(float) * msize, cudaMemcpyDeviceToHost);
	cudaMemcpy(epsilonImg, epsilonGrad, sizeof(float) * msize, cudaMemcpyDeviceToHost);
	cudaMemcpy(deltaImg,   deltaGrad,   sizeof(float) * msize, cudaMemcpyDeviceToHost);

}

void prop2D::resetTmpArray() {
	cudaMemset(prev, 0, msize * 2 * sizeof(float));
	cudaMemset(curr, 0, msize * 2 * sizeof(float));
	cudaMemset(next, 0, msize * 2 * sizeof(float));
	cudaMemset(prevR, 0, msize * 2 * sizeof(float));
	cudaMemset(currR, 0, msize * 2 * sizeof(float));
	cudaMemset(nextR, 0, msize * 2 * sizeof(float));
}

}
