#include "SEPMYinc.h"
#include "Vector.h"
#include "utilGENESIS.h"
#include "randomBound.cu"
#include "PropVTI2dGPU.cu"

using namespace std;
using namespace SEPMY;

/*
 *  code to generate RTM 2d image for GENESIS dataset with extended offset.
 *  VTI wave equation is applied on GPU
 *
 *  Input:
 *  	seed: int, used to shuffle shot gather, for quick debugging
 *  	offmin: float, minimum offset used for RTM
 *  	offmax: float, maximum offset used for RTM
 *
 *  Output:
 *
 *  Author: Yinbin Ma
 *  First created: 2018-01-23
 *  Last modified: 2018-01-23
 */

int main (int argc, char *argv[]) {
	initpar(argc, argv);

	// parameters related to machine
	int randomseed = 0; // used to shuffle the shot gather
	int deviceID = 0;   // id of GPU device
	float offmin = 100000., offmax = -1000000.; // range of offset
    int shift = 0;
    int odd = 0;

	// import parameter from command line
	getch("seed", "d", &randomseed);
	getch("offmin", "f", &offmin);
	getch("offmax", "f", &offmax);
	getch("deviceID", "d", &deviceID);
    getch("shift", "d", &shift);
    getch("odd", "d", &odd);
	cout << "offset range from " << offmin << " to " << offmax << " meters." << endl;
	cudaSetDevice(shift + odd * 4);

	// geometry for RTM
	// create space: geometry, (7, nr, ns)
	//               model,    (nx, nz)
	//               data,     (nt, nr, ns)
	//               wave,     (2, nz, nx)
	int ns, nr, nt, nz, nx, nhx = 1;
	float ot, oz, ox, dt, dz, dx;
	hyperCube realmodelCube, dataCube, geomCube;
	realmodelCube.readAxes("slow");
	dataCube.readAxes("data");
	geomCube.readAxes("head");
	hyperCube modelCube = expandModelCube(realmodelCube);
	float tmpf0, tmpf1;
	extractPar(&modelCube, nx, ox, dx, 1);
	extractPar(&modelCube, nz, oz, dz, 2);
	extractPar(&dataCube, nt, ot, dt, 1);
	extractPar(&geomCube, ns, tmpf0, tmpf1, 3);
	extractPar(&geomCube, nr, tmpf0, tmpf1, 2);

	/*
	cout << "Small cube" << endl;
	realmodelCube.axesInfo();
	cout << "Large cube" << endl;
	modelCube.axesInfo();
	cout << "Header cube" << endl;
	geomCube.axesInfo();
	cout << "Data cube" << endl;
	dataCube.axesInfo();
	*/

	// physical parameters for RTM
    float vmax = 5.0;
    float vmin = 1.5;
    float fmax = 50;
    float dtProp = 0.8;
    cout << "CFL        = " << vmax * dtProp / dx << endl;
    cout << "Dispersion = " << vmin * 1000 / fmax / dz << endl;

	// debug mode if necessary

    // Load wavelet, data and model
    float *source = new float[nt * 2 + 1];
    //sreed("time", source, sizeof(float) * (nt * 2 + 1));
	for (int it = 0; it < nt * 2 + 1; it++) {
		int x = it - nt;
		if (x == 0)
			source[it] = 1.;
		else {
			float wx = x * dt * 0.001 * 400.;
			source[it] = sin(wx) / wx;
		}
	}
//	srite("test", source, sizeof(float) * (2 * nt + 1));
//	axis ax_t(nt * 2 + 1, -dt * nt * 0.001, dt * 0.001, "shot", "sx id");
//	hyperCube tmpCube({ax_t});
//	tmpCube.writeAxes("test");
//	return 0;
    int modelSizeSmall = realmodelCube.getSize();
    int modelSize = modelCube.getSize();
    int dataSize = dataCube.getSize();
    int geomSize = geomCube.getSize();
    float *data = new float[dataSize];
    float *geom = new float[geomSize];
    sreed("data", data, sizeof(float) * dataSize);
    sreed("head", geom, sizeof(float) * geomSize);
    float *velTmp   = new float[modelSizeSmall];
    float *etaTmp   = new float[modelSizeSmall];
    float *deltaTmp = new float[modelSizeSmall];
    float *vel      = new float[modelSize];
    float *eta      = new float[modelSize];
    float *delta    = new float[modelSize];
    float *epsilon  = new float[modelSize];
    sreed("slow", velTmp, sizeof(float) * modelSizeSmall);
    sreed("eta", etaTmp, sizeof(float) * modelSizeSmall);
    sreed("delta", deltaTmp, sizeof(float) * modelSizeSmall);
    for (int i = 0; i < modelSizeSmall; i++) {
    	velTmp[i] = 1.0 / velTmp[i];
    	etaTmp[i] = max(0.0, etaTmp[i]);
    }
    fillModel(realmodelCube, modelCube, velTmp, vel);
    fillModel(realmodelCube, modelCube, etaTmp, eta);
    fillModel(realmodelCube, modelCube, deltaTmp, delta);
	for (int i = 0; i < modelSize; i++) {
		epsilon[i] = eta[i] * (1 + 2 * delta[i]) + delta[i];
	}
    float oxIn = realmodelCube.getAxis(2).o;
    float oxnew = modelCube.getAxis(1).o;
    float oxFar = realmodelCube.getAxis(2).n * dx + oxIn - dx;
    cout << "ox In = " << oxIn << ", ox Far = " << oxFar;
    cout << ", shift = " << (oxIn - oxnew) / dx << endl;


    // build the propagator
    prop2D * waveProp = new prop2D(nt, nx, nz, dt, dx, dz, ot, ox, oz, dtProp, 1, nr, nhx);
    waveProp->setModel(vel, epsilon, delta);
    waveProp->getInfo();
  	float * savedRandomV = new float[2136 * 792 * 100];
  	sreed("randV", savedRandomV, sizeof(float) * 2136 * 792 * 100);
    float *randomV = new float[nx * nz];

    // build image space
    hyperCube imgCube = buildImageCube(modelCube, nhx, ns / 4);
  	float *img = new float[nx * nz * nhx];
  	float *velGrad = new float[nx * nz];
  	float *epsilonGrad = new float[nx * nz];
  	float *deltaGrad = new float[nx * nz];
  	float *tmps = new float[nx * nz * nhx];
  	float *tmps2 = new float[nx * nz * nhx];
  	memset(img, 0, sizeof(float) * nx * nz * nhx);
  	memset(tmps, 0, sizeof(float) * nx * nz * nhx);
    // shuffle shot
    int startS = shift;
    int endS   = ns;
    int incS   = 8;
    vector<int> myvector;
    // set some values:
    for (int i=startS; i< endS; i = i + incS)
    	myvector.push_back(i); // 1 2 3 4 5 6 7 8 9
  	for (int i = 0; i < shift + 2; i++) {
  		random_shuffle (myvector.begin(), myvector.end());
  	}

	// actual run
    for(int i = 0; i < myvector.size(); i++) {
    	cout << "==============================================================================" << endl;
    	// get shot id
    	int is = myvector[i];
    	time_t t0;
    	time (&t0);
    	float sx = geom[is * nr * 7];
    	float sz = geom[is * nr * 7 + 2];

    	// valid shot
    	float lowRx = 20000000.;
    	float highRx = 0.;
    	int count = 0;
    	for (int ir = 0; ir < nr; ir++) {
    		if (geom[is * nr * 7 + ir * 7 + 6]) {
    			float off = fabs(sx - geom[is * nr * 7 + ir * 7+ 3]);
    			if (off < offmin || off > offmax)
    				geom[is * nr * 7 + ir * 7 + 6] = 0;
    		}
    		if (geom[is * nr * 7 + ir * 7 + 6]) {
    			lowRx = min(lowRx, geom[is * nr * 7 + ir * 7+ 3]);
    			highRx = max(highRx, geom[is * nr * 7 + ir * 7 + 3]);
    			count++;
    		}
    	}
    	cout << "run " << i << ", shot " << is << ", active = " << count << ", over " << nr << ",  sx = " << geom[is * nr * 7] -oxIn << ", rx from " << lowRx - oxIn << ", to " << highRx - oxIn << "should range 0 to " << oxFar - oxIn<< endl;
    	if (highRx < oxIn) {
    		cout << "    receivers not in range, continue..." << endl;
            srite("img", img, sizeof(float) * nhx * nz * nx);
            imgCube.writeAxes("img");
    		continue;
    	}
    	if (sx > oxFar) {
    		cout << "    source not in range, continue..." << endl;
            srite("img", img, sizeof(float) * nhx * nz * nx);
            imgCube.writeAxes("img");
    		continue;
    	}

    	// Actuall RTM
    	memcpy(vel, &savedRandomV[(is % 100) * nz * nx], sizeof(float) * nz * nx);
    	//memcpy(randomV, vel, sizeof(float) * nz * nx);
    	waveProp->setModel(vel, epsilon, delta);
    	waveProp->setShot(source, &sx, &sz, 1);
        waveProp->setData(&data[is * nt * nr], &geom[is * nr * 7], nr, oxIn);
        //waveProp->rtmOper(img);
        waveProp->gradOper(velGrad, epsilonGrad, deltaGrad);

        // output image
        for (int ihx = 0; ihx < nhx; ihx++) {
        	for (int iz = 2; iz < nz - 2; iz++) {
        		for (int ix = 2; ix < nx - 2; ix++) {
        			int loc = ihx * nz * nx + iz * nx + ix;
        			float scale = max(0, iz - 80);
        			//tmps[loc] = (4. * img[loc] - img[loc - 1] - img[loc + 1] - img[loc - nx] - img[loc + nx]) * scale * scale;
        			//tmps[loc] = img[loc] * scale;
        			tmps[loc] = velGrad[loc];// * scale;
        			tmps2[loc] = epsilonGrad[loc];// * scale;
        			//tmps[loc] = iz > 80 ? 1 : 0;
        		}
        	}
        }
        srite("img", tmps, sizeof(float) * nhx * nz * nx);
        imgCube.writeAxes("img");
        srite("velGrad", tmps, sizeof(float) * nz * nx);
        imgCube.writeAxes("velGrad");
        srite("epsilonGrad", tmps2, sizeof(float) * nz * nx);
        imgCube.writeAxes("epsilonGrad");
      	//memset(img, 0, sizeof(float) * nx * nz * nhx);
        // summarize the run time
    	time_t t1;
    	time (&t1);
        cout << "Processing shot " << is << ", sx = " << sx - oxIn << ", sz = " << sz;
    	cout << ", takes " << difftime (t1, t0) << " seconds." << endl;

    }

	return 0;
}
