#include "SEPMYinc.h"
#include "Vector.h"
#include "utilGENESIS.h"
#include "PropVTI2dGPU.cu"
#define Ft2M 0.3048
#define M2Ft 3.28084
using namespace std;
using namespace SEPMY;

/*
 *  code to generate RTM 2d image for GENESIS dataset with extended offset.
 *  VTI wave equation is applied on GPU
 *
 *  Input:
 *  	seed: int, used to shuffle shot gather, for quick debugging
 *  	offmin: float, minimum offset used for RTM
 *  	offmax: float, maximum offset used for RTM
 *
 *  Output:
 *
 *  Author: Yinbin Ma
 *  First created: 2018-01-23
 *  Last modified: 2018-01-23
 */

int main (int argc, char *argv[]) {
	initpar(argc, argv);

	// parameters related to machine
	int randomseed = 0; // used to shuffle the shot gather
	int deviceID = 0;   // id of GPU device
	float offmin = 100000., offmax = -1000000.; // range of offset
    int shift = 0;
    int odd = 0;

	// import parameter from command line
	getch("seed", "d", &randomseed);
	getch("offmin", "f", &offmin);
	getch("offmax", "f", &offmax);
	getch("deviceID", "d", &deviceID);
    getch("shift", "d", &shift);
    getch("odd", "d", &odd);
	cout << "offset range from " << offmin << " to " << offmax << " meters." << endl;
	cudaSetDevice(shift + odd);

	// geometry for RTM
	// create space: geometry, (7, nr, ns)
	//               model,    (nx, nz)
	//               data,     (nt, nr, ns)
	//               wave,     (2, nz, nx)
	// set 2000 as origin
	int ns, nr, noff, nt, nz, nx, nhx = 1;
	float os, ds, ooff, doff, ot, oz, ox, dt, dz, dx;
	float oxID, dxID;
	float xMin, xMax;
	float sz = 0;
	float grid = 12.5;
	nt = 1500; ot = 0.;  dt = 4.;
	getch("nmid", "d", &ns); getch("omid", "f", &os);     getch("dmid", "f", &ds);
	getch("noff", "d", &noff); getch("ooff", "f", &ooff);   getch("doff", "f", &doff);
	getch("n1", "d", &nz);   getch("o1", "f", &oz);       getch("d1", "f", &dz);
	getch("n2", "d", &nx);   getch("o2", "f", &oxID);     getch("d2", "f", &dxID);
	ox = (oxID - 2000) * grid; dx = dz;
	xMin = ox; xMax = ox + (nx - 1) * dx;
	nr = noff;
	// set up axis
	axis ax_z(nz, oz, dz, "depth", "meter");
	axis ax_x(nx, oxID, dxID, "Xline number", "unitless");
	axis ax_hx(nhx, -(nhx / 2) * SP * dx, dx * SP, "subsurface offset", "meter");
	axis ax_shot(ns / 4, 0, 1, "shot id", "unitless");
	hyperCube gradCube({ax_x, ax_z, ax_shot});
	hyperCube imgCube({ax_x, ax_z, ax_hx, ax_shot});

	// physical parameters for RTM
    float vmax = 5.0;
    float vmin = 1.5;
    float fmax = 50;
    float dtProp = 0.8;
    cout << "CFL        = " << vmax * dtProp / dx << endl;
    cout << "Dispersion = " << vmin * 1000 / fmax / dz << endl;

	// debug mode if necessary

    // Load wavelet, data and model
    float *source = new float[nt * 2 + 1];
    memset(source, 0, sizeof(float) * (nt * 2 + 1));
	//source[nt] = 1.;
	generateRicker(source, 25, dt, nt * 2 + 1, dt * nt);

	// load model
	int modelSize = nx * nz;
    float *vel      = new float[modelSize];
    float *eta      = new float[modelSize];
    float *delta    = new float[modelSize];
    float *epsilon  = new float[modelSize];
    sreed("slow",  vel,   sizeof(float) * modelSize);
    sreed("eta",   eta,   sizeof(float) * modelSize);
    sreed("delta", delta, sizeof(float) * modelSize);
    for (int i = 0; i < modelSize; i++) {
    	vel[i] = 1.0 / vel[i];
    	eta[i] = max(0.0, eta[i]);
		epsilon[i] = eta[i] * (1 + 2 * delta[i]) + delta[i];
    }

    // build the propagator
    prop2D * waveProp = new prop2D(nt, nx, nz, dt, dx, dz, ot, ox, oz, dtProp, 1, nr, nhx);
    waveProp->setModel(vel, epsilon, delta);
    waveProp->getInfo();




	//
	//
  	// float * savedRandomV = new float[modelSize * 100];
  	// sreed("bigV", savedRandomV, sizeof(float) * modelSize * 100);
	//
    // // allocate memory for imaging
  	// float *img = new float[modelSize * nhx];   memset(img, 0, sizeof(float) * modelSize * nhx);
  	// float *tmps = new float[modelSize * nhx];  memset(tmps, 0, sizeof(float) * modelSize * nhx);
  	// float *velGrad     = new float[modelSize];
  	// float *epsilonGrad = new float[modelSize];
  	// float *deltaGrad   = new float[modelSize];
	//
    // // shuffle shot
    // int startS = shift;
    // int endS   = ns;
    // int incS   = 4;
    // vector<int> myvector;
    // // set some values:
    // for (int i=startS; i< endS; i = i + incS)
    // 	myvector.push_back(i); // 1 2 3 4 5 6 7 8 9
  	// for (int i = 0; i < shift + 2; i++) {
  	// 	random_shuffle (myvector.begin(), myvector.end());
  	// }
	//
	// // actual run
    // for(int i = 0; i < myvector.size(); i++) {
    // 	cout << "==============================================================================" << endl;
    // 	// get shot id
    // 	int is = myvector[i];
    // 	time_t t0;
    // 	time (&t0);
	//
	// 	float sx = (is + os - 2000) * grid;
	// 	float sz = 0.;
	//
	// 	// create geometry
	// 	float *thisGeom = &geom[is * nt * 7];
	// 	memset(thisGeom, 0, sizeof(float) * noff * 7);
	// 	int count = noff;
	// 	float lowRx = 20000000.;
	// 	float highRx = 0.;
	// 	for (int ioff = 0; ioff < noff; ioff++) {
	// 		float offset = (ooff + ioff * doff);
	// 		float rx = sx + offset;
	// 		thisGeom[ioff * 7] = sx;
	// 		thisGeom[ioff * 7 + 3] = rx;
	// 		thisGeom[ioff * 7 + 6] = 1.;
	// 		//validate
	// 		offset = fabs(offset);
	// 		if (offset < offmin || offset > offmax || rx < xMin || rx > xMax) {
	// 			thisGeom[ioff * 7 + 6] = 0.;
	// 			count--;
	// 		} else {
	// 			lowRx = min(lowRx, rx);
	// 			highRx = max(highRx, rx);
	// 		}
	// 	}
    // 	cout << "run " << i << ", shot " << is << ", active = " << count << ", over " << noff << ",  ";
	// 	cout << "sx = " << sx / grid + 2000 << ", ";
	// 	cout << "rx from " << lowRx / grid + 2000 << ", to " << highRx / grid + 2000 << endl;
	// 	//continue;
	//
    // 	// Actuall RTM
    // 	memcpy(vel, &savedRandomV[(is % 100) * modelSize], sizeof(float) * modelSize);
    // 	waveProp->setModel(vel, epsilon, delta);
    // 	waveProp->setShot(source, &sx, &sz, 1);
    //     waveProp->setData(&data[is * nt * noff], thisGeom, noff, ox);
    //     waveProp->rtmOper(img);
    //     //waveProp->gradOper(velGrad, epsilonGrad, deltaGrad);
	// 	srite("img", img, sizeof(float) * nhx * nz * nx);
    //     imgCube.writeAxes("img");
    //     // srite("velGrad", velGrad, sizeof(float) * nz * nx);
    //     // gradCube.writeAxes("velGrad");
    //     // srite("epsilonGrad", epsilonGrad, sizeof(float) * nz * nx);
    //     // gradCube.writeAxes("epsilonGrad");
    //     // srite("deltaGrad", deltaGrad, sizeof(float) * nz * nx);
    //     // gradCube.writeAxes("deltaGrad");
    // 	time_t t1;
    // 	time (&t1);
    //     cout << "Processing shot " << is << ", sx = " << sx - ox << ", sz = " << sz;
    // 	cout << ", takes " << difftime (t1, t0) << " seconds." << endl;
	// 	//break;
	//
    // }

	return 0;
}
